function plot_solid(v, ms, ns, ls, rho)
% Plot volumetric mapping result
% 
% v: m x n x l x 3 where
%   v(:,:,:,1) are the x-coordinates
%   v(:,:,:,2) are the y-coordinates
%   v(:,:,:,3) are the z-coordinates
% ms: m spacing for the cross sections 
% ns: n spacing for the cross sections 
% ls: l spacing for the cross sections 
% rho: quantity defined on vertices
%
% If you use this code in your own work, please cite the following paper:
% [1] G. P. T. Choi and C. H. Rycroft,
%     "Volumetric density-equalizing reference map with applications."
%     Journal of Scientific Computing, 86(3), 41, 2021.
% 
% Copyright (c) 2021-2023, Gary P. T. Choi

FaceAlpha = 0.2;

M = size(v,1);
N = size(v,2);
L = size(v,3);

figure;
hold on;

% plot the slices in the first dimension
for i = [1:ms:M-ms,M]
    x = reshape(v(i,:,:,1),N,L);
    y = reshape(v(i,:,:,2),N,L);
    z = reshape(v(i,:,:,3),N,L);
    r = reshape(rho(i,:,:),N,L);

    surf(x,y,z,r,'EdgeColor','none','FaceAlpha',FaceAlpha);

    % plot the boundary of the slices
    plot3(x(:,1), y(:,1), z(:,1), 'k-', 'LineWidth', 0.5);
    plot3(x(:,end), y(:,end), z(:,end), 'k-', 'LineWidth', 0.5);
    plot3(x(1,:), y(1,:), z(1,:), 'k-', 'LineWidth', 0.5);
    plot3(x(end,:), y(end,:), z(end,:), 'k-', 'LineWidth', 0.5);

end

% plot the slices in the second dimension
for j = [1:ns:N-ns,N]
    x = reshape(v(:,j,:,1),M,L);
    y = reshape(v(:,j,:,2),M,L);
    z = reshape(v(:,j,:,3),M,L);
    r = reshape(rho(:,j,:),M,L);

    surf(x,y,z,r,'EdgeColor','none','FaceAlpha',FaceAlpha);

    % plot the boundary of the slices
    plot3(x(:,1), y(:,1), z(:,1), 'k-', 'LineWidth', 0.5);
    plot3(x(:,end), y(:,end), z(:,end), 'k-', 'LineWidth', 0.5);
    plot3(x(1,:), y(1,:), z(1,:), 'k-', 'LineWidth', 0.5);
    plot3(x(end,:), y(end,:), z(end,:), 'k-', 'LineWidth', 0.5);
end

% plot the slices in the third dimension
for k = [1:ls:L-ls,L]
    x = reshape(v(:,:,k,1),M,N);
    y = reshape(v(:,:,k,2),M,N);
    z = reshape(v(:,:,k,3),M,N);
    r = reshape(rho(:,:,k),M,N);

    surf(x,y,z,r,'EdgeColor','none','FaceAlpha',FaceAlpha);

    % plot the boundary of the slices
    plot3(x(:,1), y(:,1), z(:,1), 'k-', 'LineWidth', 0.5);
    plot3(x(:,end), y(:,end), z(:,end), 'k-', 'LineWidth', 0.5);
    plot3(x(1,:), y(1,:), z(1,:), 'k-', 'LineWidth', 0.5);
    plot3(x(end,:), y(end,:), z(end,:), 'k-', 'LineWidth', 0.5);
end

colorbar
axis equal tight
