// Volumetric density-equalizing reference map (VDERM) with the free boundary condition
//
// Usage:
// g++ -I ./eigen-3.4.0/ -Ofast -fopenmp vderm_free.cpp -o vderm_free
// ./vderm_free
//
// If you use this code in your own work, please cite the following paper:
// [1] G. P. T. Choi and C. H. Rycroft,
//    "Volumetric density-equalizing reference map with applications."
//    Journal of Scientific Computing, 86(3), 41, 2021.
//
// Copyright (c) 2021-2023, Gary P. T. Choi

#include <cstdio>
#include <cmath>
#include "omp.h"
#include <fstream>
#include <Eigen/Sparse>
#include <Eigen/IterativeLinearSolvers> 	

using namespace std;
using namespace Eigen;

const double pi = 3.14159265358979323846;

int main() {
	const int M = 48;
	const int N = 48;
	const int L = 48;
	
	double h = 1;			 // grid size
	int max_num_iter = 10000; // max number of iterations allowed in the density-equalzation
    VectorXd rho(M*N*L);     // density
    VectorXd rho_new(M*N*L); // the new density
    VectorXd map(M*N*L*3);   // the coordinates of the nodes after the deformation
    VectorXd cx(M*N*L*3);    // for upwind scheme
    VectorXd cy(M*N*L*3);    // for upwind scheme 
    VectorXd cz(M*N*L*3);    // for upwind scheme
	
	double count = 0;
	
	// define density: eight regions
	#pragma omp parallel for	
	for(int i=M/6; i<5*M/6; i++){
		for(int j=N/6; j<5*N/6; j++){
			for(int k=L/6; k<5*L/6; k++){
				if(i<M/2 && j < N/2 && k<L/2) rho(i*N*L+j*L+k) = 1;
				if(i>=M/2 && j < N/2 && k<L/2) rho(i*N*L+j*L+k) = 3;
				if(i<M/2 && j >= N/2 && k<L/2) rho(i*N*L+j*L+k) = 5;
				if(i>=M/2 && j >= N/2 && k<L/2) rho(i*N*L+j*L+k) = 7;
				if(i<M/2 && j < N/2 && k>=L/2) rho(i*N*L+j*L+k) = 9;
				if(i<M/2 && j >= N/2 && k>=L/2) rho(i*N*L+j*L+k) = 11;
				if(i>=M/2 && j < N/2 && k>=L/2) rho(i*N*L+j*L+k) = 13;
				if(i>=M/2 && j >= N/2 && k>=L/2) rho(i*N*L+j*L+k) = 15;
				#pragma omp atomic
				count += 1;
			}	
		}
	}

//	// define density: peaks
//	#pragma omp parallel for	
//	for(int i=M/6; i<5*M/6; i++){
//		for(int j=N/6; j<5*N/6; j++){
//			for(int k=L/6; k<5*L/6; k++){
//			rho(i*N*L+j*L+k) = 10+9.99*sin((i-M/6)*pi/(M*2/3)*4)*cos((j-N/6)*pi/(N*2/3)*2)*cos((k-L/6)*pi/(L*2/3)*2); // note the *3/2 because of the enlarged size
//			#pragma omp atomic
//			count += 1;
//			}
//		}
//	}
	
	// set the density at the sea
	double rho_mean = rho.sum()/count;
	#pragma omp parallel for	
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
			if(i<M/6||i>=5*M/6||j<N/6||j>=5*N/6||k<L/6||k>=5*L/6) rho(i*N*L+j*L+k) = rho_mean;
			}
		}
	}
	
	
// determine maximum stepsize (based on the CFL condition)
	VectorXd vxabs(M*N*L);
	VectorXd vyabs(M*N*L);
	VectorXd vzabs(M*N*L);
	#pragma omp parallel for
	for(int i=M/6; i<5*M/6; i++){
		for(int j=N/6; j<5*N/6; j++){
			for(int k=L/6; k<5*L/6; k++){
					if(i!=M/6 && i!=5*M/6-1){
						// \nabla_x rho / rho
	                    vxabs(i*N*L+j*L+k) = abs((rho((i+1)*N*L+j*L+k)-rho((i-1)*N*L+j*L+k))/(2*h)/rho(i*N*L+j*L+k));
	                }
	                else vxabs(i*N*L+j*L+k) = 0;
	                
	                if(j!=N/6 && j!=5*N/6-1){
	                    vyabs(i*N*L+j*L+k) = abs((rho(i*N*L+(j+1)*L+k)-rho(i*N*L+(j-1)*L+k))/(2*h)/rho(i*N*L+j*L+k));
					}
	                else vyabs(i*N*L+j*L+k) = 0;
	                
	                if(k!=L/6 && k!=5*L/6-1){
	                    vzabs(i*N*L+j*L+k) = abs((rho(i*N*L+j*L+(k+1))-rho(i*N*L+j*L+(k-1)))/(2*h)/rho(i*N*L+j*L+k));
					}
	                else vzabs(i*N*L+j*L+k) = 0;
			}
		}
	}
	VectorXd vabs(M*N*L);
	vabs = vxabs+vyabs+vzabs;
	double dt = 2*h/(3*vabs.maxCoeff());   	
	double nu = dt/h/h/h;
	printf("Timestep = 2h/(3max(|v_x|+|v_y|+|v_z|) = %.4f\n", dt);
	
	// initialize the map
	#pragma omp parallel for	
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
				map(i*N*L*3+j*L*3+k*3+0) = i*h;
				map(i*N*L*3+j*L*3+k*3+1) = j*h;
				map(i*N*L*3+j*L*3+k*3+2) = k*h;
			}	
		}
	}
	
	// write the initial state
	FILE* initial_map = fopen("results/vderm_free_initial_map.txt","w");
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
					fprintf(initial_map, "%.6f %.6f %.6f \n", map(i*N*L*3+j*L*3+k*3+0), map(i*N*L*3+j*L*3+k*3+1), map(i*N*L*3+j*L*3+k*3+2));
			}	
		}
	}
	fclose(initial_map);
	
	FILE* initial_rho = fopen("results/vderm_free_initial_rho.txt","w");
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
					fprintf(initial_rho, "%.6f \n", rho(i*N*L+j*L+k));
			}	
		}
	}
	fclose(initial_rho);
		
	FILE* size = fopen("results/vderm_free_size.txt","w");
	fprintf(size, "%d %d %d \n", M, N, L);
	fclose(size);
  
    double t0=omp_get_wtime(),t1; // get current time
  
	// create A = (I-dt*L)
  	typedef Eigen::Triplet<double> T;
	std::vector<T> tripletList;
	tripletList.reserve(7*M*N*L);
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
				// ghost node approach for the no-flux condition
				double d = i*N*L+j*L+k;
				double top, bottom, left, right, up, down;
				
				// top
				if(i==0) top = d;
				else top = (i-1)*N*L+j*L+k;
				
				// bottom
				if(i==M-1) bottom = d;
				else bottom = (i+1)*N*L+j*L+k;
				
				// left
				if(j==0) left = d;
				else left = i*N*L+(j-1)*L+k;
				
				// right
				if(j==N-1) right = d;
				else right = i*N*L+(j+1)*L+k;
				
				// up
				if(k==0) up = d;
				else up = i*N*L+j*L+(k-1);
				
				// down
				if(k==L-1) down = d;
				else down = i*N*L+j*L+(k+1);
				
				tripletList.push_back(T(d,top,-nu));
				tripletList.push_back(T(d,bottom,-nu));
				tripletList.push_back(T(d,left,-nu));
				tripletList.push_back(T(d,right,-nu));
				tripletList.push_back(T(d,up,-nu));
				tripletList.push_back(T(d,down,-nu));
				tripletList.push_back(T(d,d,1+6*nu));
			}
		}
	}
    SparseMatrix<double> A(M*N*L,M*N*L);
	A.setFromTriplets(tripletList.begin(), tripletList.end());
	ConjugateGradient<SparseMatrix<double> > solver;
	solver.compute(A);
	
    VectorXd rho_diff;
    rho_diff = rho;
    int n = 0;
    
	// Iterations for density-equalization
	while(n<max_num_iter && rho_diff.norm()/rho.mean()>1e-2){
		// update rho using implicit Euler
		rho_new = solver.solve(rho);

		// update map using the advection equation  
		#pragma omp parallel for	
		for(int i=0; i<M; i++){
			for(int j=0; j<N; j++){
				for(int k=0; k<L; k++){
					if(i!=0 && i!=M-1){
						// \nabla_x rho / rho
	                    double vx = -(rho_new((i+1)*N*L+j*L+k)-rho_new((i-1)*N*L+j*L+k))/(2*h)/rho_new(i*N*L+j*L+k); // central difference
	                     
						if(i!=1 && i!=M-2){
						// second-order upwind method for the advection equation              
	                    if(vx>0){
	                    	cx(i*N*L*3+j*L*3+k*3+0) = -vx*(3*map(i*N*L*3+j*L*3+k*3+0)-4*map((i-1)*N*L*3+j*L*3+k*3+0)+map((i-2)*N*L*3+j*L*3+k*3+0))/(2*h); 
	                    	cx(i*N*L*3+j*L*3+k*3+1) = -vx*(3*map(i*N*L*3+j*L*3+k*3+1)-4*map((i-1)*N*L*3+j*L*3+k*3+1)+map((i-2)*N*L*3+j*L*3+k*3+1))/(2*h); 
	                    	cx(i*N*L*3+j*L*3+k*3+2) = -vx*(3*map(i*N*L*3+j*L*3+k*3+2)-4*map((i-1)*N*L*3+j*L*3+k*3+2)+map((i-2)*N*L*3+j*L*3+k*3+2))/(2*h); 
						}
	                    else{
	                    	cx(i*N*L*3+j*L*3+k*3+0) = -vx*(-map((i+2)*N*L*3+j*L*3+k*3+0)+4*map((i+1)*N*L*3+j*L*3+k*3+0)-3*map(i*N*L*3+j*L*3+k*3+0))/(2*h); 
	                    	cx(i*N*L*3+j*L*3+k*3+1) = -vx*(-map((i+2)*N*L*3+j*L*3+k*3+1)+4*map((i+1)*N*L*3+j*L*3+k*3+1)-3*map(i*N*L*3+j*L*3+k*3+1))/(2*h); 
	                    	cx(i*N*L*3+j*L*3+k*3+2) = -vx*(-map((i+2)*N*L*3+j*L*3+k*3+2)+4*map((i+1)*N*L*3+j*L*3+k*3+2)-3*map(i*N*L*3+j*L*3+k*3+2))/(2*h); 
						} 
						}            
						else{
	                    // first-order upwind method for the advection equation 
	                    if(vx>0){
	                    	cx(i*N*L*3+j*L*3+k*3+0) = -vx*(map(i*N*L*3+j*L*3+k*3+0)-map((i-1)*N*L*3+j*L*3+k*3+0))/h; 
	                    	cx(i*N*L*3+j*L*3+k*3+1) = -vx*(map(i*N*L*3+j*L*3+k*3+1)-map((i-1)*N*L*3+j*L*3+k*3+1))/h; 
	                    	cx(i*N*L*3+j*L*3+k*3+2) = -vx*(map(i*N*L*3+j*L*3+k*3+2)-map((i-1)*N*L*3+j*L*3+k*3+2))/h; 
						}
	                    else{
	                    	cx(i*N*L*3+j*L*3+k*3+0) = -vx*(map((i+1)*N*L*3+j*L*3+k*3+0)-map(i*N*L*3+j*L*3+k*3+0))/h; 
	                    	cx(i*N*L*3+j*L*3+k*3+1) = -vx*(map((i+1)*N*L*3+j*L*3+k*3+1)-map(i*N*L*3+j*L*3+k*3+1))/h; 
	                    	cx(i*N*L*3+j*L*3+k*3+2) = -vx*(map((i+1)*N*L*3+j*L*3+k*3+2)-map(i*N*L*3+j*L*3+k*3+2))/h; 
						} 
						}
					}
					
					if(j!=0 && j!=N-1){
						// \nabla_y rho / rho
	                    double vy = -(rho_new(i*N*L+(j+1)*L+k)-rho_new(i*N*L+(j-1)*L+k))/(2*h)/rho_new(i*N*L+j*L+k); // central difference
	                    
						if(j!=1 && j!=N-2){
						// second-order upwind method
						if(vy>0){
	                    	cy(i*N*L*3+j*L*3+k*3+0) = -vy*(3*map(i*N*L*3+j*L*3+k*3+0)-4*map(i*N*L*3+(j-1)*L*3+k*3+0)+map(i*N*L*3+(j-2)*L*3+k*3+0))/(2*h); 
	                    	cy(i*N*L*3+j*L*3+k*3+1) = -vy*(3*map(i*N*L*3+j*L*3+k*3+1)-4*map(i*N*L*3+(j-1)*L*3+k*3+1)+map(i*N*L*3+(j-2)*L*3+k*3+1))/(2*h); 
	                    	cy(i*N*L*3+j*L*3+k*3+2) = -vy*(3*map(i*N*L*3+j*L*3+k*3+2)-4*map(i*N*L*3+(j-1)*L*3+k*3+2)+map(i*N*L*3+(j-2)*L*3+k*3+2))/(2*h); 
						} 
	                    else{
	                    	cy(i*N*L*3+j*L*3+k*3+0) = -vy*(-map(i*N*L*3+(j+2)*L*3+k*3+0)+4*map(i*N*L*3+(j+1)*L*3+k*3+0)-3*map(i*N*L*3+j*L*3+k*3+0))/(2*h); 
	                    	cy(i*N*L*3+j*L*3+k*3+1) = -vy*(-map(i*N*L*3+(j+2)*L*3+k*3+1)+4*map(i*N*L*3+(j+1)*L*3+k*3+1)-3*map(i*N*L*3+j*L*3+k*3+1))/(2*h); 
	                    	cy(i*N*L*3+j*L*3+k*3+2) = -vy*(-map(i*N*L*3+(j+2)*L*3+k*3+2)+4*map(i*N*L*3+(j+1)*L*3+k*3+2)-3*map(i*N*L*3+j*L*3+k*3+2))/(2*h); 
						} 	
						}
						else{
	                    // first-order upwind method           
	                    if(vy>0){
	                    	cy(i*N*L*3+j*L*3+k*3+0) = -vy*(map(i*N*L*3+j*L*3+k*3+0)-map(i*N*L*3+(j-1)*L*3+k*3+0))/h; 
	                    	cy(i*N*L*3+j*L*3+k*3+1) = -vy*(map(i*N*L*3+j*L*3+k*3+1)-map(i*N*L*3+(j-1)*L*3+k*3+1))/h; 
	                    	cy(i*N*L*3+j*L*3+k*3+2) = -vy*(map(i*N*L*3+j*L*3+k*3+2)-map(i*N*L*3+(j-1)*L*3+k*3+2))/h; 
						} 
	                    else{
	                    	cy(i*N*L*3+j*L*3+k*3+0) = -vy*(map(i*N*L*3+(j+1)*L*3+k*3+0)-map(i*N*L*3+j*L*3+k*3+0))/h; 
	                    	cy(i*N*L*3+j*L*3+k*3+1) = -vy*(map(i*N*L*3+(j+1)*L*3+k*3+1)-map(i*N*L*3+j*L*3+k*3+1))/h; 
	                    	cy(i*N*L*3+j*L*3+k*3+2) = -vy*(map(i*N*L*3+(j+1)*L*3+k*3+2)-map(i*N*L*3+j*L*3+k*3+2))/h; 
						} 
						}
					}
					
					
					if(k!=0 && k!=L-1){
						// \nabla_z rho / rho
	                    double vz = -(rho_new(i*N*L+j*L+(k+1))-rho_new(i*N*L+j*L+(k-1)))/(2*h)/rho_new(i*N*L+j*L+k); // central difference
	                    
						if(k!=1 && k!=L-2){
	                    // first-order upwind method            
	                    if(vz>0){
	                    	cz(i*N*L*3+j*L*3+k*3+0) = -vz*(3*map(i*N*L*3+j*L*3+k*3+0)-4*map(i*N*L*3+j*L*3+(k-1)*3+0)+map(i*N*L*3+j*L*3+(k-2)*3+0))/(2*h); 
	                    	cz(i*N*L*3+j*L*3+k*3+1) = -vz*(3*map(i*N*L*3+j*L*3+k*3+1)-4*map(i*N*L*3+j*L*3+(k-1)*3+1)+map(i*N*L*3+j*L*3+(k-2)*3+1))/(2*h); 
	                    	cz(i*N*L*3+j*L*3+k*3+2) = -vz*(3*map(i*N*L*3+j*L*3+k*3+2)-4*map(i*N*L*3+j*L*3+(k-1)*3+2)+map(i*N*L*3+j*L*3+(k-2)*3+2))/(2*h); 
						} 
	                    else{
	                    	cz(i*N*L*3+j*L*3+k*3+0) = -vz*(-map(i*N*L*3+j*L*3+(k+2)*3+0)+4*map(i*N*L*3+j*L*3+(k+1)*3+0)-3*map(i*N*L*3+j*L*3+k*3+0))/(2*h);  
	                    	cz(i*N*L*3+j*L*3+k*3+1) = -vz*(-map(i*N*L*3+j*L*3+(k+2)*3+1)+4*map(i*N*L*3+j*L*3+(k+1)*3+1)-3*map(i*N*L*3+j*L*3+k*3+1))/(2*h); 
	                    	cz(i*N*L*3+j*L*3+k*3+2) = -vz*(-map(i*N*L*3+j*L*3+(k+2)*3+2)+4*map(i*N*L*3+j*L*3+(k+1)*3+2)-3*map(i*N*L*3+j*L*3+k*3+2))/(2*h); 
						} 
						}
						else{
	                    // first-order upwind method            
	                    if(vz>0){
	                    	cz(i*N*L*3+j*L*3+k*3+0) = -vz*(map(i*N*L*3+j*L*3+k*3+0)-map(i*N*L*3+j*L*3+(k-1)*3+0))/h; 
	                    	cz(i*N*L*3+j*L*3+k*3+1) = -vz*(map(i*N*L*3+j*L*3+k*3+1)-map(i*N*L*3+j*L*3+(k-1)*3+1))/h; 
	                    	cz(i*N*L*3+j*L*3+k*3+2) = -vz*(map(i*N*L*3+j*L*3+k*3+2)-map(i*N*L*3+j*L*3+(k-1)*3+2))/h; 
						} 
	                    else{
	                    	cz(i*N*L*3+j*L*3+k*3+0) = -vz*(map(i*N*L*3+j*L*3+(k+1)*3+0)-map(i*N*L*3+j*L*3+k*3+0))/h; 
	                    	cz(i*N*L*3+j*L*3+k*3+1) = -vz*(map(i*N*L*3+j*L*3+(k+1)*3+1)-map(i*N*L*3+j*L*3+k*3+1))/h; 
	                    	cz(i*N*L*3+j*L*3+k*3+2) = -vz*(map(i*N*L*3+j*L*3+(k+1)*3+2)-map(i*N*L*3+j*L*3+k*3+2))/h; 
						} 
						}
					}
				}	
			}
		}
		map = map+dt*(cx+cy+cz);
		
		
		rho_diff = rho-rho_new;
		t1=omp_get_wtime()-t0; // count time
	    printf("Iteration %d: Time = %.4f s, ||rho_new - rho||/mean(rho)= %.4f\n", n, t1, rho_diff.norm()/rho.mean());
	    
		rho = rho_new;
		
		n++;
	}
	
	// write the final result
	FILE* final_map = fopen("results/vderm_free_final_map.txt","w");
	for(int i=0; i<M; i++){
		for(int j=0; j<N; j++){
			for(int k=0; k<L; k++){
					fprintf(final_map, "%.6f %.6f %.6f \n", map(i*N*L*3+j*L*3+k*3+0), map(i*N*L*3+j*L*3+k*3+1), map(i*N*L*3+j*L*3+k*3+2));
			}	
		}
	}
	fclose(final_map);
	
}




